/* 
 * Copyright (C) 2001-2002 the xine project
 * 	Heavily modified by Bastien Nocera <hadess@hadess.net>
 * 	Modified more by John Ellis <johne@bellatlantic.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * $Id: gtkxine.c,v 1.1.2.11 2002/09/04 04:39:29 johne Exp $
 *
 * the xine engine in a widget - implementation
 */

#include <config.h>

#include "intl.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* system */
#include <math.h>
#include <pthread.h>
/* X11 */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
/* gtk+ */
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gtk/gtkwidget.h>
#include <gtk/gtkwindow.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkinvisible.h>
/* xine */
#include <xine/video_out_x11.h>

#include "gtkxine.h"
#include "gtkxine-marshal.h"
#include "gtkxine-support.h"

/* credits */
#include <gtk/gtkbutton.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkframe.h>
#include <gtk/gtkimage.h>
#include "gtkxine-logo.xpm"


#define DEFAULT_WIDTH 320
#define DEFAULT_HEIGHT 240
#define CONFIG_FILE ".xine/config"


#define BLACK_PIXEL \
	BlackPixel ((gtx->priv->display ? gtx->priv->display : gdk_display), \
			gtx->priv->screen)

/* missing stuff from X includes */
#ifndef XShmGetEventBase
extern int XShmGetEventBase (Display *);
#endif

/* this struct is used to decouple signals coming out of the Xine threads */
typedef struct {
	gint type;		/* one of the signals in the following enum */
	GtkXineError error_type;
	char *message;		/* or NULL */
} GtkXineSignal;

/* Signals */
enum {
	ERROR,
	MOUSE_MOTION,
	EOS,
	LAST_SIGNAL
};

/* Arguments */
enum {
	PROP_0,
	PROP_FULLSCREEN,
	PROP_SPEED,
	PROP_POSITION,
	PROP_AUDIOCHANNEL,
	PROP_CURRENT_TIME,
	PROP_STREAM_LENGTH,
	PROP_PLAYING,
	PROP_SEEKABLE,
	PROP_SHOWCURSOR,
};

struct GtkXinePrivate {
	/* Xine stuff */
	xine_t *xine;
	config_values_t *config;
	vo_driver_t *vo_driver;
	ao_driver_t *ao_driver;

	/* X stuff */
	Display *display;
	int screen;
	Window video_window;
	GC gc;

	/* Other stuff */
	pthread_t thread;
	int completion_event;
	int mixer;
	int xpos, ypos;
	gboolean init_finished;

	GAsyncQueue *queue;

	/* fullscreen stuff */
	gboolean fullscreen_mode;
	int fullscreen_width, fullscreen_height;
	Window fullscreen_window, toplevel;
	Cursor no_cursor;
	GtkWidget *invisible;
	gboolean cursor_shown;

	/* config override stuff */
	gchar *video_plugin;
	gchar *audio_plugin;

	gchar *logo_mrl;
};


static void gtk_xine_class_init(GtkXineClass *klass);
static void gtk_xine_instance_init(GtkXine *gtx);

static void gtk_xine_set_property(GObject *object, guint property_id,
				  const GValue *value, GParamSpec *pspec);
static void gtk_xine_get_property(GObject *object, guint property_id,
				  GValue *value, GParamSpec *pspec);

static void gtk_xine_realize(GtkWidget *widget);
static void gtk_xine_unrealize(GtkWidget *widget);
static void gtk_xine_finalize (GObject *object);

static gint gtk_xine_expose(GtkWidget *widget, GdkEventExpose *event);

static void gtk_xine_size_allocate(GtkWidget * widget, GtkAllocation * allocation);

static GtkWidgetClass *parent_class = NULL;

static void xine_event(void *user_data, xine_event_t *event);
static void codec_reporting(void *user_data, int codec_type,
			    uint32_t fourcc, char *description, int handled);

static int gtx_table_signals[LAST_SIGNAL] = { 0 };


GtkType
gtk_xine_get_type(void)
{
	static GtkType gtk_xine_type = 0;

	if (!gtk_xine_type)
       		{
		static const GTypeInfo gtk_xine_info = {
			sizeof (GtkXineClass),
			(GBaseInitFunc) NULL,
			(GBaseFinalizeFunc) NULL,
			(GClassInitFunc) gtk_xine_class_init,
			(GClassFinalizeFunc) NULL,
			NULL /* class_data */,
			sizeof (GtkXine),
			0 /* n_preallocs */,
			(GInstanceInitFunc) gtk_xine_instance_init,
			};

		gtk_xine_type = g_type_register_static (GTK_TYPE_WIDGET,
				"GtkXine", &gtk_xine_info, (GTypeFlags)0);
		}

	return gtk_xine_type;
}

static void
gtk_xine_class_init(GtkXineClass * klass)
{

	GObjectClass *object_class;
	GtkWidgetClass *widget_class;

	object_class = (GObjectClass *) klass;
	widget_class = (GtkWidgetClass *) klass;

	parent_class = gtk_type_class (gtk_widget_get_type ());

	/* GtkWidget */
	widget_class->realize = gtk_xine_realize;
	widget_class->unrealize = gtk_xine_unrealize;
	widget_class->size_allocate = gtk_xine_size_allocate;
	widget_class->expose_event = gtk_xine_expose;

	/* GObject */
	object_class->set_property = gtk_xine_set_property;
	object_class->get_property = gtk_xine_get_property;
	object_class->finalize = gtk_xine_finalize;

	/* Properties */
	g_object_class_install_property(object_class, PROP_FULLSCREEN,
					g_param_spec_boolean("fullscreen", NULL, NULL,
					       		     FALSE, G_PARAM_READWRITE));
	g_object_class_install_property(object_class, PROP_SPEED,
					g_param_spec_int("speed", NULL, NULL,
					       		 SPEED_PAUSE, SPEED_FAST_4,
							 0, G_PARAM_READWRITE));
	g_object_class_install_property(object_class, PROP_POSITION,
					g_param_spec_int("position", NULL, NULL,
					       		 0, 65535, 0, G_PARAM_READABLE));
	g_object_class_install_property(object_class, PROP_AUDIOCHANNEL,
					g_param_spec_int("audiochannel", NULL, NULL,
					       		 0, 65535, 0, G_PARAM_READWRITE));
	g_object_class_install_property(object_class, PROP_STREAM_LENGTH,
					g_param_spec_int("stream_length", NULL, NULL,
							 0, 65535, 0, G_PARAM_READABLE));
	g_object_class_install_property(object_class, PROP_PLAYING,
					g_param_spec_boolean("playing", NULL, NULL,
							     FALSE, G_PARAM_READABLE));
	g_object_class_install_property(object_class, PROP_SEEKABLE,
					g_param_spec_boolean("seekable", NULL, NULL,
							     FALSE, G_PARAM_READABLE));
	g_object_class_install_property(object_class, PROP_SHOWCURSOR,
					g_param_spec_boolean("showcursor", NULL, NULL,
							     FALSE, G_PARAM_READWRITE));

	/* Signals */
	gtx_table_signals[ERROR] =
		g_signal_new("error",
			     G_TYPE_FROM_CLASS (object_class),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET (GtkXineClass, error),
			     NULL, NULL,
			     gtkxine_marshal_VOID__INT_STRING,
			     G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_STRING);

	gtx_table_signals[MOUSE_MOTION] =
		g_signal_new("mouse-motion",
			     G_TYPE_FROM_CLASS (object_class),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET (GtkXineClass, mouse_motion),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);

	gtx_table_signals[EOS] =
		g_signal_new("eos",
			      G_TYPE_FROM_CLASS (object_class),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (GtkXineClass, eos),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);

	if (!g_thread_supported())
		{
		g_thread_init(NULL);
		}

	gdk_threads_init();
}

static void
gtk_xine_instance_init(GtkXine * gtx)
{
	gtx->widget.requisition.width = DEFAULT_HEIGHT;
	gtx->widget.requisition.height = DEFAULT_WIDTH;

	gtx->priv = g_new0 (GtkXinePrivate, 1);
	gtx->priv->config = NULL;
	gtx->priv->xine = NULL;
	gtx->priv->vo_driver = NULL;
	gtx->priv->ao_driver = NULL;
	gtx->priv->display = NULL;
	gtx->priv->fullscreen_mode = FALSE;
	gtx->priv->init_finished = FALSE;
	gtx->priv->mixer = -1;
	gtx->priv->cursor_shown = TRUE;

	gtx->priv->queue = g_async_queue_new();

	gtx->priv->video_plugin = NULL;
	gtx->priv->audio_plugin = NULL;

	gtx->priv->logo_mrl = NULL;
}

static void
gtk_xine_finalize(GObject *object)
{
	GtkXine *gtx = (GtkXine *)object;

	/* FIXME check this! */

	g_free(gtx->priv->logo_mrl);
	g_free(gtx->priv->video_plugin);
	g_free(gtx->priv->audio_plugin);

	g_free(gtx->priv);

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
dest_size_cb(void *gtx_gen,
	     int video_width, int video_height,
	     int *dest_width, int *dest_height)
{
	GtkXine *gtx = (GtkXine *) gtx_gen;

	if (gtx->priv->fullscreen_mode)
       		{
		*dest_width = gtx->priv->fullscreen_width;
		*dest_height = gtx->priv->fullscreen_height;
		}
       	else
       		{
		*dest_width = gtx->widget.allocation.width;
		*dest_height = gtx->widget.allocation.height;
		}
}

static void
frame_output_cb(void *gtx_gen,
		int video_width, int video_height,
		int *dest_x, int *dest_y,
		int *dest_width, int *dest_height, int *win_x, int *win_y)
{
	GtkXine *gtx = (GtkXine *) gtx_gen;

	*dest_x = 0;
	*dest_y = 0;
	*win_x = gtx->priv->xpos;
	*win_y = gtx->priv->ypos;
	if (gtx->priv->fullscreen_mode)
       		{
		*dest_width = gtx->priv->fullscreen_width;
		*dest_height = gtx->priv->fullscreen_height;
		}
       	else
       		{
		*dest_width = gtx->widget.allocation.width;
		*dest_height = gtx->widget.allocation.height;
		}
}

static vo_driver_t *
load_video_out_driver(GtkXine *gtx)
{
	double res_h, res_v;
	x11_visual_t vis;
	char **driver_ids;
	int i;
	const char *video_driver_id;
	vo_driver_t *vo_driver;

	vis.display = gtx->priv->display;
	vis.screen = gtx->priv->screen;
	vis.d = gtx->priv->video_window;
	res_h = (DisplayWidth (gtx->priv->display, gtx->priv->screen) * 1000 /
		 DisplayWidthMM (gtx->priv->display, gtx->priv->screen));
	res_v = (DisplayHeight (gtx->priv->display, gtx->priv->screen) * 1000 /
		 DisplayHeightMM (gtx->priv->display, gtx->priv->screen));
	vis.display_ratio = res_h / res_v;

	if (fabs (vis.display_ratio - 1.0) < 0.01)
       		{
		vis.display_ratio = 1.0;
		}

	vis.dest_size_cb = dest_size_cb;
	vis.frame_output_cb = frame_output_cb;
	vis.user_data = gtx;

	/* attempt preferred driver */
	if (gtx->priv->video_plugin)
		{
		vo_driver = xine_load_video_output_plugin(gtx->priv->config, gtx->priv->video_plugin,
							  VISUAL_TYPE_X11, (void *) &vis);
		if (vo_driver) return vo_driver;

		printf("Failed to load preferred video plugin: %s, using auto detection...\n", gtx->priv->video_plugin);
		}

	/* Try to init video with stored information */
	video_driver_id = gtx->priv->config->register_string(gtx->priv->config, "video.driver", "auto",
							     "video driver to use", NULL, NULL, NULL);
	if (strcmp(video_driver_id, "auto") == 0)
       		{
		vo_driver = xine_load_video_output_plugin(gtx->priv->config, "auto",
							  VISUAL_TYPE_X11, (void *) &vis);

		if (vo_driver) return vo_driver;
		}

	/* use specific driver */
	driver_ids = xine_list_video_output_plugins(VISUAL_TYPE_X11);	
	i = 0;
	vo_driver = NULL;
	while (driver_ids[i] && !vo_driver)
		{
		if (strcmp(video_driver_id, driver_ids[i]) == 0)
			{
			vo_driver = xine_load_video_output_plugin(gtx->priv->config, driver_ids[i],
								  VISUAL_TYPE_X11, (void *) &vis);
			}
		if (vo_driver)
	       		{
			gtx->priv->config->update_string(gtx->priv->config, "video.driver", driver_ids[i]);
			}
		i++;
		}

	if (!vo_driver)
		{
		printf("gtkxine warning: video plugin %s not found or failed to load\n", video_driver_id);
		}
	
	g_strfreev(driver_ids);

	return vo_driver;
}

static ao_driver_t *
load_audio_out_driver(GtkXine * gtx)
{
	ao_driver_t *ao_driver = NULL;
	char **driver_ids;
	const char *audio_driver_id;
	int i;

	/* attemp preferred driver */
	if (gtx->priv->audio_plugin)
		{
		ao_driver = xine_load_audio_output_plugin(gtx->priv->config, gtx->priv->audio_plugin);
		if (ao_driver) return ao_driver;

		printf("Failed to load preferred audio plugin: %s, using auto detection...\n", gtx->priv->audio_plugin);
		}
	
	/* try to init audio with stored information */
	audio_driver_id = gtx->priv->config->register_string(gtx->priv->config, "audio.driver", "auto",
							     "audio driver to use",
							     NULL, NULL, NULL);
	if (strcmp(audio_driver_id, "auto") == 0)
		{
		ao_driver =  xine_load_audio_output_plugin(gtx->priv->config, (char *)audio_driver_id);

		if (ao_driver) return ao_driver;
		}

	driver_ids = xine_list_audio_output_plugins();
	i = 0;
	ao_driver = NULL;
	while (driver_ids[i] && !ao_driver)
		{
		if (strcmp(audio_driver_id, driver_ids[i]) == 0)
			{
			ao_driver = xine_load_audio_output_plugin (gtx->priv->config, driver_ids[i]);
			}

		if (ao_driver)
	       		{
			gtx->priv->config->update_string(gtx->priv-> config, "audio.driver", driver_ids[i]);
			}
		i++;
		}

	if (!ao_driver)
		{
		printf("gtkxine warning: audio plugin %s not found or failed to load,\nusing null driver\n", audio_driver_id);
		ao_driver = xine_load_audio_output_plugin(gtx->priv->config, "null");
		}

	g_strfreev(driver_ids);

	return ao_driver;
}

static gboolean
x_window_is_visible(Display *display, Window window)
{
	XWindowAttributes  wattr;
	Status status;

	if(display == NULL || window == None) return FALSE;

	XLockDisplay(display);
	status = XGetWindowAttributes(display, window, &wattr);
	XUnlockDisplay(display);

	if(status != BadDrawable &&
	   status != BadWindow &&
	   wattr.map_state == IsViewable)
		{
		return TRUE;
		}

	return FALSE;
}

static void *
xine_thread(void *gtx_gen)
{
	GtkXine *gtx = (GtkXine *)gtx_gen;
	XEvent event;

	gtx->priv->init_finished = TRUE;

	while (1)
       		{
		XNextEvent(gtx->priv->display, &event);

		switch (event.type)
			{
			case Expose:
				if (event.xexpose.count != 0)
					{
					break;
					}
				gtx->priv->vo_driver->gui_data_exchange(gtx->priv->vo_driver,
									GUI_DATA_EX_EXPOSE_EVENT, &event);
				break;
			case FocusIn:
				/* happens only in fullscreen mode,
				 * wait for window to get visible to avoid BadMatch
				 */
				while (!x_window_is_visible(gtx->priv->display, gtx->priv->toplevel))
					{
					usleep(5000);
					}

				XSetInputFocus(gtx->priv->display,
					       gtx->priv->toplevel, RevertToNone,
					       CurrentTime);
				break;
			}

		if (event.type == gtx->priv->completion_event)
			{
			gtx->priv->vo_driver->gui_data_exchange(gtx->priv->vo_driver,
								GUI_DATA_EX_COMPLETION_EVENT, &event);
			}
		}

	pthread_exit(NULL);
	return NULL;
}

static gboolean
configure_cb(GtkWidget *widget,
	     GdkEventConfigure *event, gpointer user_data)
{
	GtkXine *gtx = (GtkXine *)user_data;;

	gtx->priv->xpos = event->x;
	gtx->priv->ypos = event->y;

	return FALSE;
}

static unsigned char bm_no_data[] = { 0, 0, 0, 0, 0, 0, 0, 0 };

static void
gtk_xine_realize(GtkWidget *widget)
{
	GtkXine *gtx;
	XGCValues values;
	Pixmap bm_no;
	char *configfile;

	g_return_if_fail(widget != NULL);
	g_return_if_fail(GTK_IS_XINE (widget));

	gtx = GTK_XINE(widget);

	/* set realized flag */
	GTK_WIDGET_SET_FLAGS(widget, GTK_REALIZED);

	/* create our own video window */
	gtx->priv->video_window = XCreateSimpleWindow(gdk_display,
			GDK_WINDOW_XWINDOW(gtk_widget_get_parent_window(widget)),
			0, 0,
			widget->allocation.width, widget->allocation.height,
		       	1, BLACK_PIXEL, BLACK_PIXEL);

	widget->window = gdk_window_foreign_new(gtx->priv->video_window);

	/* prepare for fullscreen playback */
	gtx->priv->fullscreen_width = DisplayWidth (gdk_display, gdk_x11_get_default_screen());
	gtx->priv->fullscreen_height = DisplayHeight (gdk_display, gdk_x11_get_default_screen());

	gtx->priv->toplevel = GDK_WINDOW_XWINDOW (gdk_window_get_toplevel(gtk_widget_get_parent_window(widget)));

	/* track configure events of toplevel window */
	g_signal_connect(GTK_OBJECT(gtk_widget_get_toplevel (widget)), "configure-event",
			 GTK_SIGNAL_FUNC(configure_cb), gtx);

#ifdef DEBUG_GTKXINE
	printf("xine_thread: init threads\n");
#endif

	if (!XInitThreads())
       		{
		g_signal_emit(G_OBJECT(gtx),
			      gtx_table_signals[ERROR], 0, 0,
			      _("Could not initialise the threads support.\n"
			      "You should install a thread-safe Xlib."));
		return;
		}

#ifdef DEBUG_GTKXINE
	printf("xine_thread: open display\n");
#endif

	gtx->priv->display = XOpenDisplay(NULL);

	if (!gtx->priv->display)
       		{
		g_signal_emit(G_OBJECT(gtx),
			      gtx_table_signals[ERROR], 1,
			      _("XOpenDisplay failed"));
		return;
		}

	XLockDisplay(gtx->priv->display);

	gtx->priv->screen = DefaultScreen(gtx->priv->display);

	if (XShmQueryExtension(gtx->priv->display) == True)
       		{
		gtx->priv->completion_event = XShmGetEventBase (gtx->priv->display) + ShmCompletion;
		}
       	else
       		{
		gtx->priv->completion_event = -1;
		}

	XSelectInput(gtx->priv->display, gtx->priv->video_window,
		     StructureNotifyMask | ExposureMask |
		     ButtonPressMask | PointerMotionMask);

	/* generate and init a config "object" */
	configfile = g_build_path (G_DIR_SEPARATOR_S, g_get_home_dir(), CONFIG_FILE, NULL);
	gtx->priv->config = xine_config_file_init(configfile);
	g_free(configfile);

	/* setup demuxer stragety */
	gtx->priv->config->register_string(gtx->priv->config, "misc.demux_strategy", "reverse",
					   "demuxer selection strategy",
					   "{ default  reverse  content  extension }, default: 0",
					   NULL, NULL);
	gtx->priv->config->update_string (gtx->priv->config, "misc.demux_strategy", "reverse");

	/* set a custom logo */
	if (gtx->priv->logo_mrl)
		{
		gtx->priv->config->register_string(gtx->priv->config, "misc.logo_mrl", gtx->priv->logo_mrl,
			       			   "logo", NULL, NULL, NULL);
		gtx->priv->config->update_string(gtx->priv->config, "misc.logo_mrl", gtx->priv->logo_mrl);
		}

	/* load audio, video drivers */
	gtx->priv->vo_driver = load_video_out_driver(gtx);

	if (!gtx->priv->vo_driver)
		{
		g_signal_emit(G_OBJECT(gtx),
			      gtx_table_signals[ERROR], 0,
			      GTX_STARTUP,
			      _("Could not find a suitable video output."));
		return;
		}

	gtx->priv->ao_driver = load_audio_out_driver(gtx);

	/* init xine engine */
	gtx->priv->xine = xine_init(gtx->priv->vo_driver, gtx->priv->ao_driver, gtx->priv->config);

	values.foreground = BLACK_PIXEL;
	values.background = BLACK_PIXEL;

	gtx->priv->gc = XCreateGC(gtx->priv->display, gtx->priv->video_window,
				  (GCForeground | GCBackground), &values);

	XUnlockDisplay (gtx->priv->display);

	/* create mouse cursors */
	bm_no = XCreateBitmapFromData(gtx->priv->display,
				      gtx->priv->video_window,
				      bm_no_data, 8, 8);
	gtx->priv->no_cursor = XCreatePixmapCursor(gtx->priv->display, bm_no, bm_no,
						   (XColor *) & BLACK_PIXEL,
						   (XColor *) & BLACK_PIXEL, 0, 0);

	/* Setup xine events and codec reporting */
	xine_register_event_listener(gtx->priv->xine, xine_event, (void *)gtx);
	xine_register_report_codec_cb(gtx->priv->xine, codec_reporting, (void *)gtx);

	/* screen saver handler */
	scrsaver_init(gtx->priv->display);

	/* now, create a xine thread */
	pthread_create(&gtx->priv->thread, NULL, xine_thread, gtx);
}

static gboolean
gtk_xine_idle_signal(GtkXine *gtx)
{
	GtkXineSignal *signal;
	int queue_length;

	signal = g_async_queue_try_pop(gtx->priv->queue);
	if (!signal) return FALSE;


	gdk_threads_enter();

	switch(signal->type)
		{
		case ERROR:
			g_signal_emit(G_OBJECT (gtx),
				      gtx_table_signals[ERROR], 0,
				      signal->error_type, signal->message);
			break;
		case MOUSE_MOTION:
			/* mouse motion doesn't need to go through here */
			break;
		case EOS:
			g_signal_emit(G_OBJECT (gtx),
				      gtx_table_signals[EOS], 0, NULL);
			break;
		default:
			break;
		}

	g_free(signal->message);
	g_free(signal);

	queue_length = g_async_queue_length (gtx->priv->queue);

	gdk_threads_leave();

	return (queue_length > 0);
}

static void
xine_event (void *user_data, xine_event_t *event)
{
	GtkXine *gtx = (GtkXine *)user_data;

	/* XINE_EVENT_MOUSE_MOVE doesn't actually work */
	if (event->type == XINE_EVENT_PLAYBACK_FINISHED)
		{
		GtkXineSignal *signal;

		signal = g_new0(GtkXineSignal, 1);
		signal->type = EOS;
		g_async_queue_push(gtx->priv->queue, signal);
		g_idle_add((GSourceFunc)gtk_xine_idle_signal, gtx);
		}
}

static void
codec_reporting(void *user_data, int codec_type,
		uint32_t fourcc, char *description, int handled)
{
	GtkXine *gtx = (GtkXine *)user_data;
	char fourcc_txt[10];

	/* store fourcc as text */
	*(uint32_t *)fourcc_txt = fourcc;
	fourcc_txt[4] = '\0';

	if (!handled)
		{
		if (codec_type == XINE_CODEC_VIDEO)
			{
			GtkXineSignal *signal;

			signal = g_new0(GtkXineSignal, 1);
			signal->type = ERROR;

			/* display fourcc if no description available */
			if (!description[0]) description = fourcc_txt;
			signal->error_type = GTX_NO_CODEC;
			signal->message = g_strdup_printf(_("No video plugin available to decode '%s'."), description);
			g_async_queue_push(gtx->priv->queue, signal);
			g_idle_add((GSourceFunc)gtk_xine_idle_signal, gtx);
			}
		}
}

static void
xine_error(GtkXine *gtx)
{
	GtkXineSignal *signal;
	int error;

	error = xine_get_error(gtx->priv->xine);
	if (error == XINE_ERROR_NONE) return;

	signal = g_new0(GtkXineSignal, 1);
	signal->type = ERROR;

	switch (error)
		{
		case XINE_ERROR_NO_INPUT_PLUGIN:
			signal->error_type = GTX_NO_INPUT_PLUGIN;
			break;
		case XINE_ERROR_NO_DEMUXER_PLUGIN:
			signal->error_type = GTX_NO_DEMUXER_PLUGIN;
			break;
		case XINE_ERROR_DEMUXER_FAILED:
			signal->error_type = GTX_DEMUXER_FAILED;
			break;
		default:
			break;
		}

	g_async_queue_push(gtx->priv->queue, signal);
	g_idle_add((GSourceFunc)gtk_xine_idle_signal, gtx);
}

static void
gtk_xine_unrealize(GtkWidget *widget)
{
	GtkXine *gtx;

	g_return_if_fail(widget != NULL);
	g_return_if_fail(GTK_IS_XINE(widget));

	gtx = GTK_XINE(widget);

	/* stop event thread */
	xine_exit(gtx->priv->xine);

	/* Hide all windows */
	if (GTK_WIDGET_MAPPED(widget))
		{
		gtk_widget_unmap(widget);
		}

	GTK_WIDGET_UNSET_FLAGS(widget, GTK_MAPPED);

	/* This destroys widget->window and unsets the realized flag */
	if (GTK_WIDGET_CLASS (parent_class)->unrealize)
		{
		(*GTK_WIDGET_CLASS (parent_class)->unrealize) (widget);
		}
}

GtkWidget *
gtk_xine_new(gint width, gint height)
{
	GtkWidget *gtx;
       
	gtx = GTK_WIDGET(gtk_type_new(gtk_xine_get_type()));

	/* defaults are fine if both are negative */
	if (width <= 0 && height <= 0) return gtx;

	/* figure out the missing measure from the other one with a 4:3 ratio */
	if (width <= 0) width = (int) (height * 4 / 3);
	if (height <= 0) height = (int) (width * 3 / 4);

	GTK_XINE (gtx)->widget.requisition.width = width;
	GTK_XINE (gtx)->widget.requisition.height = height;

	return gtx;
}

gboolean
gtk_xine_check(GtkXine *gtx)
{
	g_return_val_if_fail(gtx != NULL, FALSE);
	g_return_val_if_fail(GTK_IS_XINE(gtx), FALSE);

	if (!gtx->priv->xine) return FALSE;

	return gtx->priv->init_finished;
}

GtkWidget *
gtk_xine_new_with_plugins(gint width, gint height, const gchar *video, const gchar *audio)
{
	GtkXine *gtx;
       
	gtx = GTK_XINE(gtk_xine_new(width, height));
	gtx->priv->video_plugin = g_strdup(video);
	gtx->priv->audio_plugin = g_strdup(audio);
	return GTK_WIDGET(gtx);
}

void
gtk_xine_set_logo(GtkXine *gtx, const gchar *mrl)
{
	g_return_if_fail(gtx != NULL);
	g_return_if_fail(GTK_IS_XINE(gtx));
//	g_return_if_fail(gtx->priv->xine != NULL);

	g_free(gtx->priv->logo_mrl);
	gtx->priv->logo_mrl = g_strdup(mrl);
}

static gint
gtk_xine_expose(GtkWidget *widget, GdkEventExpose * event)
{
	return FALSE;
}

static void
gtk_xine_size_allocate(GtkWidget * widget, GtkAllocation * allocation)
{
	GtkXine *gtx;

	g_return_if_fail(widget != NULL);
	g_return_if_fail(GTK_IS_XINE (widget));

#ifdef DEBUG_GTKXINE
	printf("ALLOCATE x: %d y: %d w: %d h: %d", allocation->x,
	       allocation->y, allocation->width,
	       allocation->height);
#endif

	gtx = GTK_XINE (widget);

	widget->allocation = *allocation;
	gtx->priv->xpos = allocation->x;
	gtx->priv->ypos = allocation->y;

	if (GTK_WIDGET_REALIZED (widget))
       		{
		/* HACK it seems to be 1 pixel off, weird */
		gdk_window_move_resize(widget->window,
				       allocation->x - 1, allocation->y - 1,
				       allocation->width, allocation->height);
		}
}

/* Commands */

gint
gtk_xine_play(GtkXine * gtx, gchar * mrl, gint pos, gint start_time)
{
	int error;

	g_return_val_if_fail(gtx != NULL, -1);
	g_return_val_if_fail(GTK_IS_XINE (gtx), -1);
	g_return_val_if_fail(gtx->priv->xine != NULL, -1);

	error = xine_play(gtx->priv->xine, mrl, pos, start_time);
	if (!error)
		{
		xine_error(gtx);
		return FALSE;
		}
	return TRUE;
}

void
gtk_xine_stop(GtkXine * gtx)
{
	g_return_if_fail(gtx != NULL);
	g_return_if_fail(GTK_IS_XINE (gtx));
	g_return_if_fail(gtx->priv->xine != NULL);

	xine_stop(gtx->priv->xine);
}

/* Properties */
static void
gtk_xine_set_property(GObject *object, guint property_id,
		      const GValue *value, GParamSpec *pspec)
{
	GtkXine *gtx;

	g_return_if_fail(GTK_IS_XINE(object));

	gtx = GTK_XINE(object);

	switch (property_id)
		{
		case PROP_FULLSCREEN:
			gtk_xine_set_fullscreen(gtx, g_value_get_boolean(value), NULL);
			break;
		case PROP_SPEED:
			gtk_xine_set_speed(gtx, g_value_get_int(value));
			break;
		case PROP_AUDIOCHANNEL:
			gtk_xine_set_audio_channel(gtx, g_value_get_int(value));
			break;
		case PROP_SHOWCURSOR:
			gtk_xine_set_show_cursor(gtx, g_value_get_boolean(value));
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
			break;
		}
}

static void
gtk_xine_get_property(GObject *object, guint property_id,
		      GValue *value, GParamSpec *pspec)
{
	GtkXine *gtx;

	g_return_if_fail(GTK_IS_XINE(object));

	gtx = GTK_XINE(object);

	switch (property_id)
		{
		case PROP_FULLSCREEN:
			g_value_set_boolean(value, gtk_xine_is_fullscreen(gtx));
			break;
		case PROP_SPEED:
			g_value_set_int(value, gtk_xine_get_speed(gtx));
			break;
		case PROP_POSITION:
			g_value_set_int(value, gtk_xine_get_position(gtx));
			break;
		case PROP_AUDIOCHANNEL:
			g_value_set_int(value, gtk_xine_get_audio_channel(gtx));
			break;
		case PROP_STREAM_LENGTH:
			g_value_set_int(value, gtk_xine_get_total_time(gtx));
			break;
		case PROP_PLAYING:
			g_value_set_boolean(value, gtk_xine_is_playing(gtx));
			break;
		case PROP_SEEKABLE:
			g_value_set_boolean(value, gtk_xine_is_seekable(gtx));
			break;
		case PROP_SHOWCURSOR:
			g_value_set_boolean(value, gtk_xine_get_show_cursor(gtx));
			g_message ("set cursor");
			break;
		default:
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
			break;
		}
}

void
gtk_xine_set_speed(GtkXine * gtx, gint speed)
{

	g_return_if_fail(gtx != NULL);
	g_return_if_fail(GTK_IS_XINE (gtx));
	g_return_if_fail(gtx->priv->xine != NULL);

	xine_set_speed(gtx->priv->xine, speed);
}

gint
gtk_xine_get_speed(GtkXine * gtx)
{

	g_return_val_if_fail(gtx != NULL, SPEED_NORMAL);
	g_return_val_if_fail(GTK_IS_XINE (gtx), SPEED_NORMAL);
	g_return_val_if_fail(gtx->priv->xine != NULL, SPEED_NORMAL);

	return xine_get_speed(gtx->priv->xine);
}

gint
gtk_xine_get_position(GtkXine * gtx)
{

	g_return_val_if_fail(gtx != NULL, 0);
	g_return_val_if_fail(GTK_IS_XINE (gtx), 0);
	g_return_val_if_fail(gtx->priv->xine != NULL, 0);

	if (gtk_xine_is_playing(gtx))
		{
		return xine_get_current_position (gtx->priv->xine);
		}

	return 0;
}

void
gtk_xine_set_audio_channel(GtkXine * gtx, gint audio_channel)
{
	g_return_if_fail(gtx != NULL);
	g_return_if_fail(GTK_IS_XINE (gtx));
	g_return_if_fail(gtx->priv->xine != NULL);

	xine_select_audio_channel(gtx->priv->xine, audio_channel);
}

gint
gtk_xine_get_audio_channel(GtkXine * gtx)
{

	g_return_val_if_fail(gtx != NULL, 0);
	g_return_val_if_fail(GTK_IS_XINE (gtx), 0);
	g_return_val_if_fail(gtx->priv->xine != NULL, 0);

	return xine_get_audio_selection(gtx->priv->xine);
}


#if 0
/* FIXME: remove this */
#define MWM_HINTS_DECORATIONS   (1L << 1)
#define PROP_MWM_HINTS_ELEMENTS 5
typedef struct {
	uint32_t flags;
	uint32_t functions;
	uint32_t decorations;
	int32_t input_mode;
	uint32_t status;
} MWMHints;

#define _NET_WM_STATE_REMOVE        0   /* remove/unset property */
#define _NET_WM_STATE_ADD           1   /* add/set property */
#define _NET_WM_STATE_TOGGLE        2   /* toggle property  */

static void
wmspec_change_xwindow_state(gboolean add,
			    Window window, GdkAtom state1, GdkAtom state2)
{
	XEvent xev;

	xev.xclient.type = ClientMessage;
	xev.xclient.serial = 0;
	xev.xclient.send_event = True;
	xev.xclient.display = gdk_display;
	xev.xclient.window = window;
	xev.xclient.message_type = gdk_x11_get_xatom_by_name ("_NET_WM_STATE");
	xev.xclient.format = 32;
	xev.xclient.data.l[0] = add ? _NET_WM_STATE_ADD : _NET_WM_STATE_REMOVE;
	xev.xclient.data.l[1] = gdk_x11_atom_to_xatom (state1);
	xev.xclient.data.l[2] = gdk_x11_atom_to_xatom (state2);

	XSendEvent(gdk_display,
		   GDK_WINDOW_XID (gdk_get_default_root_window ()),
		   False,
		   SubstructureRedirectMask | SubstructureNotifyMask,
		   &xev);
}
#endif

static gboolean
motion_notify_event_cb(GtkWidget * widget, GdkEventMotion * event, gpointer user_data)
{
	GtkXine *gtx = GTK_XINE(user_data);

	g_signal_emit(G_OBJECT(gtx),
		      gtx_table_signals[MOUSE_MOTION], 0, NULL);

	return TRUE;
}

void
gtk_xine_set_fullscreen(GtkXine * gtx, gboolean fullscreen, const gchar *window_title)
{
	XSizeHints hint;
	XEvent xev;
	GdkWindow *win;

	g_return_if_fail(gtx != NULL);
	g_return_if_fail(GTK_IS_XINE (gtx));
	g_return_if_fail(gtx->priv->xine != NULL);

	if (fullscreen == gtx->priv->fullscreen_mode) return;

	if (!window_title) window_title = "GtkXine";

	gtx->priv->fullscreen_mode = fullscreen;

	XLockDisplay(gtx->priv->display);

	if (gtx->priv->fullscreen_mode)
		{
		hint.x = 0;
		hint.y = 0;
		hint.width = gtx->priv->fullscreen_width;
		hint.height = gtx->priv->fullscreen_height;

		gtx->priv->fullscreen_window = XCreateSimpleWindow(gtx->priv->display,
								   GDK_ROOT_WINDOW (),
								   0, 0,
								   gtx->priv->fullscreen_width,
								   gtx->priv->fullscreen_height, 1,
								   BLACK_PIXEL, BLACK_PIXEL);

		hint.win_gravity = StaticGravity;
		hint.flags = PPosition | PSize | PWinGravity;

		XSetStandardProperties(gtx->priv->display,
				       gtx->priv->fullscreen_window,
				       window_title, window_title, None,
				       NULL, 0, 0);

		XSetWMNormalHints(gtx->priv->display,
				  gtx->priv->fullscreen_window, &hint);

		XSelectInput(gtx->priv->display,
			     gtx->priv->fullscreen_window,
			     StructureNotifyMask | ExposureMask |
			     FocusChangeMask);

		xwindow_set_fullscreen(gtx->priv->display, gtx->priv->fullscreen_window);

		/* Map window. */
		XMapRaised(gtx->priv->display,
			   gtx->priv->fullscreen_window);

		XFlush(gtx->priv->display);

		/* Wait for map. */
		do	{
			XMaskEvent(gtx->priv->display, StructureNotifyMask, &xev);
			}
			while (xev.type != MapNotify || xev.xmap.event != gtx->priv->fullscreen_window);

		XSetInputFocus(gtx->priv->display,
			       gtx->priv->toplevel, RevertToNone,
			       CurrentTime);
		XMoveWindow(gtx->priv->display,
			    gtx->priv->fullscreen_window, 0, 0);

		gtx->priv->vo_driver->gui_data_exchange(gtx->priv->vo_driver, GUI_DATA_EX_DRAWABLE_CHANGED,
							(void *) gtx->priv->fullscreen_window);

		/* switch off mouse cursor */
		gtk_xine_set_show_cursor(gtx, FALSE);

		/* setup the mouse motion stuff */
		win = gdk_window_foreign_new(gtx->priv->fullscreen_window);
		gtx->priv->invisible = gtk_invisible_new();
		gtk_widget_show(gtx->priv->invisible);
		gdk_window_set_user_data(win, gtx->priv->invisible);
		gdk_window_set_events(win, GDK_POINTER_MOTION_MASK);

		g_signal_connect(GTK_OBJECT(gtx->priv->invisible), "motion-notify-event",
				 GTK_SIGNAL_FUNC(motion_notify_event_cb), gtx);

		scrsaver_disable(gtx->priv->display);
		}
       	else
       		{
		gtx->priv->vo_driver->gui_data_exchange(gtx->priv->vo_driver, GUI_DATA_EX_DRAWABLE_CHANGED,
							(void *) gtx->priv->video_window);

		XDestroyWindow(gtx->priv->display,
			       gtx->priv->fullscreen_window);
		gtx->priv->invisible = NULL;

		scrsaver_enable(gtx->priv->display);
		}

	XUnlockDisplay(gtx->priv->display);
}

gint
gtk_xine_is_fullscreen(GtkXine * gtx)
{

	g_return_val_if_fail(gtx != NULL, 0);
	g_return_val_if_fail(GTK_IS_XINE (gtx), 0);
	g_return_val_if_fail(gtx->priv->xine != NULL, 0);

	return gtx->priv->fullscreen_mode;
}

gboolean
gtk_xine_can_set_volume(GtkXine *gtx)
{
	static int can_set_volume = FALSE;
	int caps, mixer;

	g_return_val_if_fail(gtx != NULL, 0);
	g_return_val_if_fail(GTK_IS_XINE (gtx), 0);
	g_return_val_if_fail(gtx->priv->xine != NULL, 0);

	if (can_set_volume == TRUE) return can_set_volume;

	caps = xine_get_audio_capabilities(gtx->priv->xine);
	if (caps & AO_CAP_PCM_VOL)
		{
		mixer = AO_PROP_PCM_VOL;
		}
	else if (caps & AO_CAP_MIXER_VOL)
		{
		mixer = AO_PROP_MIXER_VOL;
		}
	else
		{
		/* FIXME this is wrong, just set mixer to default? */
		return FALSE;
		}

	if (caps & (AO_CAP_PCM_VOL | AO_CAP_MIXER_VOL))
		{
		can_set_volume = TRUE;
		}

	gtx->priv->mixer = mixer;

	return can_set_volume;
}

void
gtk_xine_set_volume(GtkXine *gtx, gint volume)
{
	g_return_if_fail(gtx != NULL);
	g_return_if_fail(GTK_IS_XINE (gtx));
	g_return_if_fail(gtx->priv->xine != NULL);

	if(gtk_xine_can_set_volume(gtx))
		{
		volume = CLAMP (volume, 0, 100);
		xine_set_audio_property(gtx->priv->xine, gtx->priv->mixer, volume);
		}
}

gint
gtk_xine_get_volume(GtkXine *gtx)
{
	int volume = 0;

	g_return_val_if_fail(gtx != NULL, 0);
	g_return_val_if_fail(GTK_IS_XINE(gtx), 0);
	g_return_val_if_fail(gtx->priv->xine != NULL, 0);

	if(gtk_xine_can_set_volume(gtx))
		{
		volume = xine_get_audio_property(gtx->priv->xine, gtx->priv->mixer);
		}

	return volume;
}

void
gtk_xine_set_show_cursor(GtkXine *gtx, gboolean show_cursor)
{
	g_return_if_fail(gtx != NULL);
	g_return_if_fail(GTK_IS_XINE (gtx));
	g_return_if_fail(gtx->priv->xine != NULL);

	if (!gtk_xine_is_fullscreen(gtx)) return;

	gdk_error_trap_push();
	if (!show_cursor)
		{
		XDefineCursor(gtx->priv->display,
			      gtx->priv->fullscreen_window,
			      gtx->priv->no_cursor);
		}
	else
		{
		XUndefineCursor(gtx->priv->display,
				gtx->priv->fullscreen_window);
		}
	XFlush(gtx->priv->display);
	gdk_flush();
	gdk_error_trap_pop();

	gtx->priv->cursor_shown = show_cursor;
}

gboolean
gtk_xine_get_show_cursor(GtkXine *gtx)
{
	g_return_val_if_fail(gtx != NULL, FALSE);
	g_return_val_if_fail(GTK_IS_XINE (gtx), FALSE);
	g_return_val_if_fail(gtx->priv->xine != NULL, FALSE);

	return gtx->priv->cursor_shown;
}

gchar **
gtk_xine_get_autoplay_plugins(GtkXine *gtx)
{

	g_return_val_if_fail(gtx != NULL, NULL);
	g_return_val_if_fail(GTK_IS_XINE (gtx), NULL);
	g_return_val_if_fail(gtx->priv->xine != NULL, NULL);

	return xine_get_autoplay_input_plugin_ids(gtx->priv->xine);
}

gint
gtk_xine_get_current_time(GtkXine *gtx)
{
	g_return_val_if_fail(gtx != NULL, 0);
	g_return_val_if_fail(GTK_IS_XINE (gtx), 0);
	g_return_val_if_fail(gtx->priv->xine != NULL, 0);

	return xine_get_current_time(gtx->priv->xine);
}

gint
gtk_xine_get_total_time(GtkXine *gtx)
{
	g_return_val_if_fail(gtx != NULL, 0);
	g_return_val_if_fail(GTK_IS_XINE (gtx), 0);
	g_return_val_if_fail(gtx->priv->xine != NULL, 0);

	return xine_get_stream_length(gtx->priv->xine);
}

gboolean
gtk_xine_is_playing(GtkXine *gtx)
{
	g_return_val_if_fail(gtx != NULL, 0);
	g_return_val_if_fail(GTK_IS_XINE (gtx), 0);
	g_return_val_if_fail(gtx->priv->xine != NULL, 0);

	return (xine_get_status(gtx->priv->xine) == XINE_PLAY);
}

gboolean
gtk_xine_is_seekable(GtkXine * gtx)
{
	g_return_val_if_fail(gtx != NULL, 0);
	g_return_val_if_fail(GTK_IS_XINE (gtx), 0);
	g_return_val_if_fail(gtx->priv->xine != NULL, 0);

	return xine_is_stream_seekable(gtx->priv->xine);
}

void
gtk_xine_toggle_aspect_ratio(GtkXine *gtx)
{
	int tmp;

	g_return_if_fail(gtx != NULL);
	g_return_if_fail(GTK_IS_XINE (gtx));
	g_return_if_fail(gtx->priv->xine != NULL);

	/* FIXME: toggle, this is merely incrementing? */
	tmp = gtx->priv->vo_driver->get_property(gtx->priv->vo_driver, VO_PROP_ASPECT_RATIO);
	gtx->priv->vo_driver->set_property(gtx->priv->vo_driver, VO_PROP_ASPECT_RATIO, tmp + 1);
}

void
gtk_xine_set_scale_ratio(GtkXine *gtx, gfloat ratio)
{
	GtkWindow *toplevel;
	uint8_t *y, *u, *v;
	int width, height, ratio_code, format;
	int new_w, new_h;

	y = u = v = NULL;

	if (gtx->priv->fullscreen_mode == TRUE) return;

	if (xine_get_current_frame(gtx->priv->xine,
				   &width, &height, &ratio_code, &format,
				   &y, &u, &v) == 0)
		{
		return;
		}

	new_w = width * ratio;
	new_h = height * ratio;

#if 0
	/* y, u and v don't need to be freed apparently */
	g_free(y);
	g_free(u);
	g_free(v);
#endif

	/* don't scale to something bigger than the screen, and leave us some room */
	if (new_w > gdk_screen_width() - 128 ||
	    new_h > gdk_screen_height() - 128) return;

	toplevel = GTK_WINDOW(gtk_widget_get_toplevel(GTK_WIDGET(gtx)));

	gtk_window_set_resizable(toplevel, FALSE);
	gtx->widget.allocation.width = new_w;
	gtx->widget.allocation.height = new_h;
	gtk_widget_set_size_request(gtk_widget_get_parent(GTK_WIDGET(gtx)), new_w, new_h);
	gtk_widget_queue_resize(gtk_widget_get_parent(GTK_WIDGET(gtx)));
	while (gtk_events_pending()) gtk_main_iteration();	/* baaaaaad..... */
	gtk_window_set_resizable (toplevel, TRUE);
}


/*-----------------
 * credits, logo
 *-----------------
 */

const gchar *gtk_xine_version_string(void)
{
	static gchar *version = NULL;

	if (!version) version = g_strdup_printf("%d.%d.%d",
						xine_get_major_version(),
						xine_get_minor_version(),
						xine_get_sub_version());
	return version;
}

const gchar **gtk_xine_logo(void)
{
	return (const gchar **)gtkxine_logo_xpm;
}

static GtkWidget *gtkxine_about_window = NULL;

static gint ui_slik_about_delete_cb(GtkWidget *widget, GdkEventAny *event, gpointer data)
{
	gtk_widget_destroy(gtkxine_about_window);
	gtkxine_about_window = NULL;
	return TRUE;
}

static void gtk_xine_about(GtkWidget *widget, gpointer data)
{
	GtkWidget *frame;
	GtkWidget *vbox;
	GtkWidget *label;
	GtkWidget *image;

	GdkPixbuf *pb;

	gchar *buf;
	
	if (gtkxine_about_window)
		{
		gdk_window_raise(gtkxine_about_window->window);
		return;
		}

	gtkxine_about_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect(GTK_OBJECT(gtkxine_about_window), "delete_event",
			 GTK_SIGNAL_FUNC(ui_slik_about_delete_cb), NULL);

	gtk_window_set_wmclass(GTK_WINDOW(gtkxine_about_window), "about", "Xine");
	gtk_window_set_title(GTK_WINDOW(gtkxine_about_window), _("About - Xine"));
	gtk_container_border_width(GTK_CONTAINER(gtkxine_about_window), 5);

	frame = gtk_frame_new(NULL);
	gtk_container_border_width(GTK_CONTAINER(frame), 5);
	gtk_container_add(GTK_CONTAINER(gtkxine_about_window), frame);
	gtk_widget_show(frame);

	vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_border_width(GTK_CONTAINER(vbox), 5);
	gtk_container_add(GTK_CONTAINER(frame), vbox);
	gtk_widget_show(vbox);

	pb = gdk_pixbuf_new_from_xpm_data((const char **)gtkxine_logo_xpm);
	image = gtk_image_new_from_pixbuf(pb);
	gdk_pixbuf_unref(pb);

	gtk_box_pack_start(GTK_BOX(vbox), image, FALSE, FALSE, 0);
	gtk_widget_show(image);

	buf = g_strdup_printf(_("XINE\nXine media library\nversion %s\nCopyright (c) 2000-2002 the xine project\nhttp://xine.sourceforge.net\n\nThis is a gtk xine widget,\noriginally based on portions of the Totem media player."), gtk_xine_version_string());
	label = gtk_label_new(buf);
	g_free(buf);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	gtk_widget_show(gtkxine_about_window);
}

GtkWidget *gtk_xine_credits(void)
{
	GtkWidget *frame;
	GtkWidget *hbox;
	GtkWidget *label;
	GtkWidget *image;

	GdkPixbuf *pb;

	gchar *buf;

	frame = gtk_button_new();
	gtk_button_set_relief(GTK_BUTTON(frame), GTK_RELIEF_NONE);
	gtk_container_border_width(GTK_CONTAINER(frame), 10);
	g_signal_connect(GTK_OBJECT(frame), "clicked",
			 GTK_SIGNAL_FUNC(gtk_xine_about), NULL);

	hbox = gtk_hbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(frame), hbox);
	gtk_widget_show(hbox);

	pb = gdk_pixbuf_new_from_xpm_data((const char **)gtkxine_logo_xpm);
	image = gtk_image_new_from_pixbuf(pb);
	gdk_pixbuf_unref(pb);

	gtk_box_pack_start(GTK_BOX(hbox), image, FALSE, FALSE, 0);
	gtk_widget_show(image);

	buf = g_strdup_printf(_("XINE inside\nversion %s"), gtk_xine_version_string());
	label = gtk_label_new(buf);
	g_free(buf);
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);

	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	return frame;
}


