/*
 * GQoob
 * (C) 2002 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */

#include "gqoob.h"
#include "io_ogg123.h"

#include "ui_fileops.h"
#include "ui_utildlg.h"

#include <fcntl.h>
#include <time.h>

#define OGG123_BINARY_INFO "ogginfo"
#define OGG123_BINARY_COMMENT "vorbiscomment"

/* ----------------------------------------------------------
   input / output interface to ogg123, note this was based
   on the GQmpeg version, but as been stripped down to only
   include the file information utilities
   ----------------------------------------------------------*/

/* options */
gint ogg123_read_info = TRUE;


static gint ogg123_info_found = FALSE;
static gint ogg123_comment_found = FALSE;


/*
 *-----------------------------------------------------------------------------
 * pipe/fork util
 *-----------------------------------------------------------------------------
 */

static gint ogg123_pipe_open(char *ogg_argv[], gint *ogg_child, int ogg_pipe[2])
{
	pid_t frk_pid;

	/* Create the pipe. */
	if (pipe(ogg_pipe))
		{
		fprintf (stderr, _("Pipe failed.\n"));
		return FALSE;
		}

	if (debug_mode) printf("ogg123_pipe_open: %s\n", ogg_argv[0]);

	/* Create the child process. */
	frk_pid = fork ();
	if (frk_pid == (pid_t) 0)
		{
		/* This is the child process. */
		dup2(ogg_pipe[1], 1);
		dup2(ogg_pipe[1], 2);
		close(ogg_pipe[0]);

		execvp(ogg_argv[0], ogg_argv);
		printf(_("unable to run %s (in the path?)\n"), ogg_argv[0]);
		_exit(1);
		}
	else if (frk_pid < (pid_t) 0)
		{
		/* The fork failed. */
		fprintf (stderr, _("Fork failed.\n"));
		close(ogg_pipe[0]);
		close(ogg_pipe[1]);
		*ogg_child = -1;
		return FALSE;
		}
	else
		{
		/* This is the parent process. */
		*ogg_child = (int) frk_pid;

		close(ogg_pipe[1]);
		ogg_pipe[1] = -1;
		}

	if (debug_mode) printf("ogg123_pipe_open pid = %d\n", *ogg_child);

	return TRUE;
}

static FILE *ogg123_pipe_open_fd(char *ogg_argv[], gint *ogg_child, int ogg_pipe[2])
{
	FILE *f;

	if (!ogg123_pipe_open(ogg_argv, ogg_child, ogg_pipe))
		{
		printf("Failed to run \"%s\n\"", ogg_argv[0]);
		return NULL;
		}
	f = fdopen(ogg_pipe[0], "r");
	if (!f)
		{
		close(ogg_pipe[0]);
		return NULL;
		}

	return f;
}

/*
 *----------------------------------------------------------------------------
 * title/description utils
 *----------------------------------------------------------------------------
 */

static const gchar *info_line_value(gchar *text)
{
	gchar *ptr, *ptr2;

	ptr = text;
	while (*ptr != '=' && *ptr != '\0') ptr++;
	if (*ptr == '\0') return NULL;
	ptr++;

	/* remove new line char */
	ptr2 = ptr;
	while (*ptr2 != '\0')
		{
		if (*ptr2 == '\n' || *ptr2 == '\r') *ptr2 = '\0';
		ptr2++;
		}

	return ptr;
}

static gint info_line_key_test(gchar *text, const gchar *key)
{
	if (!text || !key) return FALSE;
	return (strncmp(text, key, strlen(key)) == 0);
}

static gchar *info_line_key(gchar *text)
{
	gchar *ptr;
	gchar *ret;

	if (!text) return NULL;

	ptr = text;
	while (*ptr != '=' && *ptr != '\0') ptr++;
	if (*ptr == '\0') return NULL;

	*ptr = '\0';
	ret = g_strdup(text);
	*ptr = '=';

	return ret;
}

OGGInfo *ogg123_info_get(const gchar *path)
{
	OGGInfo *info = NULL;
	FILE *f;
	char *command[3];
	gint child_pid;
	int child_pipe[2];
	char buf[128];

	if (!ogg123_info_found) return NULL;

	command[0] = OGG123_BINARY_INFO;
	command[1] = (char *)path;
	command[2] = NULL;

	f = ogg123_pipe_open_fd(command, &child_pid, child_pipe);
	if (!f) return NULL;

	info = g_new0(OGGInfo, 1);

	while (fgets(buf, sizeof(buf), f) != NULL)
		{
		const gchar *value;
		value = info_line_value(buf);

		if (info_line_key_test(buf, "filename"))
			{
			info->pass_integrity = TRUE;
			}
		else if (info_line_key_test(buf, "length"))
			{
			info->length = (gint)strtol(value, NULL, 10);
			}
		else if (info_line_key_test(buf, "header_integrity") ||
			 info_line_key_test(buf, "stream_integrity"))
			{
			info->pass_integrity = info->pass_integrity & (strncmp(value, "pass", 4) == 0);
			}
		else if (info_line_key_test(buf, "bitrate_nominal"))
			{
			info->bitrate_nominal = (gint)strtol(value, NULL, 10);
			}
		else if (info_line_key_test(buf, "bitrate_average"))
			{
			info->bitrate_average = (gint)strtol(value, NULL, 10);
			}
		else if (info_line_key_test(buf, "bitrate_upper"))
			{
			if (strncmp(value, "none", 4) == 0)
				info->bitrate_upper = 0;
			else
				info->bitrate_upper = (gint)strtol(value, NULL, 10);
			}
		else if (info_line_key_test(buf, "bitrate_lower"))
			{
			if (strncmp(value, "none", 4) == 0)
				info->bitrate_lower = 0;
			else
				info->bitrate_lower = (gint)strtol(value, NULL, 10);
			}
		else if (info_line_key_test(buf, "channels"))
			{
			info->channels = (gint)strtol(value, NULL, 10);
			}
		}

	fclose(f);	/* also closes child_pipe[0] */
	waitpid(child_pid, NULL, 0);	/* eat zombie */

	info->bytes = filesize(path);

	return info;
}

/*
 * List is ordered key, value, key, value ...
 * use path_list_free() to free the returned list
 */
GList *ogg123_comment_get(const gchar *path)
{
	GList *list = NULL;
	FILE *f;
	char *command[4];
	gint child_pid;
	int child_pipe[2];
	char buf[128];

	if (!ogg123_comment_found) return NULL;

	command[0] = OGG123_BINARY_COMMENT;
	command[1] = "-l";
	command[2] = (char *)path;
	command[3] = NULL;

	f = ogg123_pipe_open_fd(command, &child_pid, child_pipe);
	if (!f) return NULL;

	while (fgets(buf, sizeof(buf), f) != NULL)
		{
		gchar *key;

		key = info_line_key(buf);

		if (key)
			{
			const gchar *value;
			value = info_line_value(buf);

			if (value)
				{
				list = g_list_prepend(list, key);
				list = g_list_prepend(list, g_strdup(value));
				key = NULL;
				}
			}
		g_free(key);
		}

	fclose(f);	/* also closes child_pipe[0] */
	waitpid(child_pid, NULL, 0);	/* eat zombie */

	return g_list_reverse(list);
}

const gchar *ogg123_comment_value(GList *list, const gchar *key)
{
	GList *work;

	if (!key) return NULL;

	work = list;
	while(work)
		{
		const gchar *poss = work->data;

		if (strcasecmp(poss, key) == 0)
			{
			if (work->next)
				{
				work = work->next;
				if (work) return (const gchar *)work->data;
				}
			return NULL;
			}

		work = work->next;
		if (work) work = work->next;
		}

	return NULL;
}

gint ogg123_comment_set(const gchar *path, GList *comments)
{
	FILE *f;
	GList *errlist;
	GList *comment_list;
	GList *work;
	char buf[128];
	int argc;
	char **argv;
	gint child_pid;
	int child_pipe[2];
	
	if (!ogg123_comment_found) return TRUE;
	if (!path) return FALSE;

	comment_list = NULL;
	errlist = NULL;

	/* get the size needed for argv */
	work = comments;
	argc = 4;
	while (work && work->next)
		{
		if (work->data) argc+=2;
		work = work->next;
		work = work->next;
		}

	argv = g_malloc0(argc * sizeof(argv[0]));
	argc = 0;
	argv[argc] = OGG123_BINARY_COMMENT;
	argc++;
	argv[argc] = "-w";
	argc++;
	argv[argc] = (char *)path;
	argc++;

	/* fill comment arguments */
	work = comments;
	while (work && work->next)
		{
		const gchar *key;
		const gchar *value;

		key = work->data;
		value = work->next->data;

		if (key && strlen(key) > 0 && value && strlen(value) > 0)
			{
			argv[argc] = "-t";
			argc++;
			argv[argc] = g_strdup_printf("%s=%s", key, value);
			/* remember allocated strings for later free'ing */
			comment_list = g_list_prepend(comment_list, argv[argc]);
			argc++;
			}

		work = work->next;
		work = work->next;
		}

	argv[argc] = NULL;

	f = ogg123_pipe_open_fd(argv, &child_pid, child_pipe);
	if (f)
		{
		while (fgets(buf, sizeof(buf), f) != NULL)
			{
			if (strlen(buf) > 0)
				{
				errlist = g_list_append(errlist, g_strdup(buf));
				}
			}
		fclose(f);
		waitpid(child_pid, NULL, 0);
		}

	if (errlist)
		{
		GenericDialog *gd;
		GtkWidget *tw;
		GtkTextIter iter;
		GtkTextBuffer *buffer;
		gchar *msg;

		msg = g_strdup_printf(_("Failed to set comments for file:\n%s"), path);
		gd = generic_dialog_new(_("Comment set failed - GQoob"), msg,
					"GQoob", "warning", TRUE, NULL, NULL);
		g_free(msg);

	        generic_dialog_add(gd, _("Ok"), NULL, TRUE);

		tw = gtk_text_view_new();
		gtk_widget_set_usize(tw, 400, 200);
		gtk_text_view_set_editable(GTK_TEXT_VIEW(tw), FALSE);
		gtk_box_pack_start(GTK_BOX(gd->vbox), tw, TRUE, TRUE, 5);
		gtk_widget_show(tw);

		buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(tw));
		gtk_text_buffer_create_tag(buffer, "monospace",
					   "family", "monospace", NULL);
		gtk_text_buffer_get_iter_at_offset(buffer, &iter, 0);

		work = errlist;
		while (work)
			{
			gchar *txt = work->data;

			gtk_text_buffer_insert_with_tags_by_name(buffer, &iter, txt, -1,
								 "monospace", NULL);
			work = work->next;
			}

		gtk_widget_show(gd->dialog);
		}

	path_list_free(errlist);
	path_list_free(comment_list);
	return (errlist == NULL);
}

/*
 *----------------------------------------------------------------------------
 * module callback funcs
 *----------------------------------------------------------------------------
 */

void ogg123_data_init(SongData *sd)
{
	sd->type_description = _("Ogg bitstream audio file");
}

gint ogg123_data_set(SongData *sd, gint read_info)
{
	if (read_info && !sd->info_loaded)
		{
		OGGInfo *info;
		GList *comments;

		sd->info_loaded = TRUE;

		if (!ogg123_read_info) return TRUE;

		info = ogg123_info_get(sd->path);
		if (info)
			{
			sd->length = info->length;
			sd->bit_rate = info->bitrate_nominal / 1000;
			sd->channels = info->channels;
			sd->bit_depth = 16;	/* these two correct ? */
			sd->khz_rate = 44;
			if (!info->pass_integrity) sd->flags |= SONG_FLAG_PLAY_FAIL;
			g_free(info);
			}

		comments = ogg123_comment_get(sd->path);

		g_free(sd->title);
		sd->title = g_strdup(ogg123_comment_value(comments, "TITLE"));
		g_free(sd->artist);
		sd->artist = g_strdup(ogg123_comment_value(comments, "ARTIST"));
		g_free(sd->album);
		sd->album = g_strdup(ogg123_comment_value(comments, "ALBUM"));
		g_free(sd->genre);
		sd->genre = g_strdup(ogg123_comment_value(comments, "GENRE"));
		g_free(sd->comment);
		sd->comment = g_strdup(ogg123_comment_value(comments, "COMMENT"));
		g_free(sd->year);
		sd->year = g_strdup(ogg123_comment_value(comments, "YEAR"));

		path_list_free(comments);
		}

	return TRUE;
}

/*
 *----------------------------------------------------------------------------
 * player module interface routines
 *----------------------------------------------------------------------------
 */

void ogg123_init(void)
{
	ogg123_info_found = file_in_path(OGG123_BINARY_INFO);
	if (!ogg123_info_found)
		{
		printf(_("\"%s\" not found, ogg info support disabled.\n"), OGG123_BINARY_INFO);
		}

	ogg123_comment_found = file_in_path(OGG123_BINARY_COMMENT);
	if (!ogg123_comment_found)
		{
		printf(_("\"%s\" not found, ogg comment support disabled.\n"), OGG123_BINARY_COMMENT);
		}
}

