/*
 * (SLIK) SimpLIstic sKin functions
 * (C) 2002 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */

#include "ui2_includes.h"
#include "ui2_typedefs.h"
#include "ui2_overlay.h"
#include "ui2_overlay_edit.h"

#include "ui2_display.h"
#include "ui2_main.h"
#include "ui2_parse.h"
#include "ui2_skin.h"
#include "ui2_util.h"
#include "ui2_widget.h"
#include "ui_pixbuf_ops.h"


typedef struct _OverlayCallbackData OverlayCallbackData;
struct _OverlayCallbackData
{
	void (*size_func)(OverlayData *od, const gchar *key, gint hidden, gpointer data);
	gpointer size_data;
};

static WidgetType type_id = -1;


/*
 *-----------------------------
 * new / free
 *-----------------------------
 */

OverlayData *overlay_new(gint x, gint y, gint width, gint height, gint dynamic)
{
	OverlayData *od;

	overlay_type_init();

	util_size(&x);
	util_size(&y);
	util_size(&width);
	util_size(&height);

	od = g_new0(OverlayData, 1);

	od->x = x;
	od->y = y;
	od->width = width;
	od->height = height;
	od->dynamic = dynamic;

	return od;
}

void overlay_free(OverlayData *od)
{
	if (!od) return;
	g_free(od);
}

static void overlay_free_cb(gpointer data)
{
	overlay_free((OverlayData *)data);
}

/*
 *-----------------------------
 * draw
 *-----------------------------
 */

static void overlay_draw(gpointer data, const gchar *key, gint update, gint force, GdkPixbuf *pb, UIData *ui)
{
	OverlayData *od = data;

	if (!od) return;

	if (force)
		{
		pixbuf_draw_rect_fill(pb, od->x, od->y, od->width, od->height, 0, 0, 0, 255);
		ui_display_render_area(ui, od->x, od->y, od->width, od->height, NULL);
		}
}

/*
 *-----------------------------
 * ui funcs
 *-----------------------------
 */

static gint overlay_get_geometry(gpointer widget, gint *x, gint *y, gint *w, gint *h)
{
	OverlayData *od = widget;

	*x = od->x;
	*y = od->y;
	*w = od->width;
	*h = od->height;

	return TRUE;
}

static void overlay_set_coord(gpointer widget, gint x, gint y)
{
	OverlayData *od = widget;

	od->x = x;
	od->y = y;
}

static void overlay_set_size(gpointer widget, gint dev_w, gint dev_h)
{
	OverlayData *od = widget;
	UIData *ui;
	OverlayCallbackData *cd;
	WidgetData *wd;

	ui = skin_get_ui(od->skin);
	if (!ui) return;

	wd = ui_widget_get_by_widget(ui, od);
	if (!wd) return;

	if (od->dynamic)
		{
		if (!wd->anchor_right)
			{
			od->width += dev_w;
			if (od->width < 1) od->width = 1;
			}
		if (!wd->anchor_bottom)
			{
			od->height += dev_h;
			if (od->height < 1) od->height = 1;
			}
		}

	cd = ui_get_registered_callbacks(ui, wd->key, type_id);
	if (cd && cd->size_func)
		{
		/* need to do this now, so that hidden flag is passed correctly */
		skin_widget_check_bounds(od->skin, wd);

		cd->size_func(od, wd->key, (wd->hidden || !wd->in_bounds), cd->size_data);
		}
}

/* in this init the inital viewport size is sent via the size_func that the app registered */
static void overlay_init(gpointer widget, const gchar *key, UIData *ui)
{
	OverlayData *od = widget;
	OverlayCallbackData *cd;
	WidgetData *wd;

	wd = ui_widget_get_by_widget(ui, od);
	if (!wd) return;

	cd = ui_get_registered_callbacks(ui, wd->key, type_id);
	if (cd && cd->size_func)
		{
		skin_widget_check_bounds(od->skin, wd);
		cd->size_func(od, wd->key, (wd->hidden || !wd->in_bounds), cd->size_data);
		}
}

static WidgetData *overlay_parse(SkinData *skin, GList *list, const gchar *skin_dir, const gchar *key, gint edit)
{
	WidgetData *wd = NULL;
	OverlayData *od;
	gint x, y;
	gint width, height;
	gint dynamic;

	/* req */
	if (!key_list_read_int(list, "x", &x)) return NULL;
	if (!key_list_read_int(list, "y", &y)) return NULL;
	if (!key_list_read_int(list, "width", &width)) return NULL;
	if (!key_list_read_int(list, "height", &height)) return NULL;

	/* opt */
	dynamic = key_list_read_bool(list, "dynamic");

	od = overlay_new(x, y, width, height, dynamic);

	if (od)
		{
		wd = overlay_register(skin, od, key, NULL);
		}

	return wd;
}


/*
 *-----------------------------
 * register ui / app side
 *-----------------------------
 */

WidgetData *overlay_register(SkinData *skin, OverlayData *od, const gchar *key, const gchar *text_id)
{
	if (od) od->skin = skin;
	return skin_register_widget(skin, key, text_id, type_id, od);
}

WidgetData *overlay_register_to_skin(const gchar *key, gint x, gint y, gint width, gint height, gint dynamic,
				     SkinData *skin, const gchar *text_id)
{
	OverlayData *od;

	od = overlay_new(x, y, width, height, dynamic);
	return overlay_register(skin, od, key, text_id);
}

RegisterData *overlay_register_key(const gchar *key, UIData *ui,
				   void (*size_func)(OverlayData *od, const gchar *key, gint hidden, gpointer data), gpointer size_data)
{
	OverlayCallbackData *cd;

	overlay_type_init();

	cd = g_new0(OverlayCallbackData, 1);

	cd->size_func = size_func;
	cd->size_data = size_data;

	return ui_register_key(ui, key, type_id, cd, sizeof(OverlayCallbackData));
}

/*
 *-----------------------------
 * app funcs
 *-----------------------------
 */


/*
 *-----------------------------
 * init
 *-----------------------------
 */

WidgetType overlay_type_id(void)
{
	return type_id;
}

void overlay_type_init(void)
{
	WidgetObjectData *od;

	if (type_id != -1) return;

	od = ui_widget_type_new("overlay");
	type_id = od->type;

	od->func_draw = overlay_draw;
	od->func_free = overlay_free_cb;

	od->func_get_geometry = overlay_get_geometry;
	od->func_set_coord = overlay_set_coord;
	od->func_set_size = overlay_set_size;

	od->func_parse = overlay_parse;
	od->func_init = overlay_init;

	overlay_type_init_edit(od);
}


